@@ -1,7 +1,7 @@
 /*
  * ProFTPD: mod_ctrls_admin -- a module implementing admin control handlers
  *
- * Copyright (c) 2000-2010 TJ Saunders
+ * Copyright (c) 2000-2013 TJ Saunders
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
  *
  * You should have received a copy of the GNU General Public License
  * along with this program; if not, write to the Free Software
- * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307, USA.
+ * Foundation, Inc., 51 Franklin Street, Suite 500, Boston, MA 02110-1335, USA.
  *
  * As a special exemption, TJ Saunders and other respective copyright holders
  * give permission to link this program with OpenSSL, and distribute the
@@ -25,14 +25,14 @@
  * This is mod_controls, contrib software for proftpd 1.2 and above.
  * For more information contact TJ Saunders <tj@castaglia.org>.
  *
- * $Id: mod_ctrls_admin.c,v 1.34.2.1 2010/01/10 21:05:45 castaglia Exp $
+ * $Id: mod_ctrls_admin.c,v 1.47 2013-10-13 22:51:36 castaglia Exp $
  */
 
 #include "conf.h"
 #include "privs.h"
 #include "mod_ctrls.h"
 
-#define MOD_CTRLS_ADMIN_VERSION		"mod_ctrls_admin/0.9.5"
+#define MOD_CTRLS_ADMIN_VERSION		"mod_ctrls_admin/0.9.7"
 
 /* Make sure the version of proftpd is as necessary. */
 #if PROFTPD_VERSION_NUMBER < 0x0001030001
@@ -62,21 +62,26 @@ static ctrls_acttab_t ctrls_admin_acttab[];
 /* Pool for this module's use */
 static pool *ctrls_admin_pool = NULL;
 
+static unsigned int ctrls_admin_nrestarts = 0;
+static time_t ctrls_admin_start = 0;
+
 /* Support routines
  */
 
 #if 0
 /* Will be used when scheduled shutdowns are supported.. */
 static unsigned char isnumeric(char *str) {
-  while (str && isspace((int) *str))
+  while (str && PR_ISSPACE(*str)) {
     str++;
+  }
 
   if (!str || !*str)
     return FALSE;
 
   for (; str && *str; str++) {
-    if (!isdigit((int) *str))
+    if (!PR_ISDIGIT(*str)) {
       return TRUE;
+    }
   }
 
   return 1;
@@ -87,9 +92,9 @@ static int respcmp(const void *a, const void *b) {
   return strcmp(*((char **) a), *((char **) b));
 }
 
+#ifdef PR_USE_DEVEL
 static pr_ctrls_t *mem_ctrl = NULL;
 
-#ifdef PR_USE_DEVEL
 static void mem_printf(const char *fmt, ...) {
   char buf[PR_TUNABLE_BUFFER_SIZE];
   va_list msg;
@@ -113,7 +118,7 @@ static int ctrls_handle_debug(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
 
   /* Check the debug ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "debug")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "debug")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -130,19 +135,31 @@ static int ctrls_handle_debug(pr_ctrls_t *ctrl, int reqargc,
   if (strcmp(reqargv[0], "level") == 0) {
     int level = 0;
 
-    if (reqargc != 2) {
-      pr_ctrls_add_response(ctrl, "debug: missing required parameters");
+    if (reqargc != 1 &&
+        reqargc != 2) {
+      pr_ctrls_add_response(ctrl, "debug: wrong number of parameters");
       return -1;
     }
 
-    if ((level = atoi(reqargv[1])) < 0) {
-      pr_ctrls_add_response(ctrl, "debug level must not be negative");
-      return -1; 
-    }
+    if (reqargc == 1) {
+      /* The user is requesting the current debug level.  Easy enough. */
+      level = pr_log_setdebuglevel(0);
+      (void) pr_log_setdebuglevel(level);
+
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: level set to %d", level);
+      pr_ctrls_add_response(ctrl, "debug level set to %d", level);
+
+    } else if (reqargc == 2) {
+      level = atoi(reqargv[1]);
+      if (level < 0) {
+        pr_ctrls_add_response(ctrl, "debug level must not be negative");
+        return -1; 
+      }
   
-    pr_log_setdebuglevel(level);
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: level set to %d", level);
-    pr_ctrls_add_response(ctrl, "debug level set to %d", level);
+      pr_log_setdebuglevel(level);
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: level set to %d", level);
+      pr_ctrls_add_response(ctrl, "debug level set to %d", level);
+    }
 
 #ifdef PR_USE_DEVEL
   /* Handle 'debug memory' requests */
@@ -158,7 +175,7 @@ static int ctrls_handle_debug(pr_ctrls_t *ctrl, int reqargc,
     pr_pool_debug_memory(mem_printf);
     mem_ctrl = NULL;
 
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: dumped memory info");
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "debug: dumped memory info");
 
 #endif /* PR_USE_DEVEL */
 
@@ -175,7 +192,7 @@ static int ctrls_handle_dns(pr_ctrls_t *ctrl, int reqargc,
   int bool;
 
   /* Check the dns ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "dns")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "dns")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -215,7 +232,7 @@ static int ctrls_handle_dns(pr_ctrls_t *ctrl, int reqargc,
 
     ServerUseReverseDNS = bool;
 
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION, "dns: UseReverseDNS set to '%s'",
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "dns: UseReverseDNS set to '%s'",
       bool ? "on" : "off");
     pr_ctrls_add_response(ctrl, "dns: UseReverseDNS set to '%s'",
       bool ? "on" : "off");
@@ -227,7 +244,7 @@ static int ctrls_handle_dns(pr_ctrls_t *ctrl, int reqargc,
 static int admin_addr_down(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
     unsigned int port) {
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "down: disabling %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "down: disabling %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   if (pr_ipbind_close(addr, port, FALSE) < 0) {
@@ -254,7 +271,7 @@ static int ctrls_handle_down(pr_ctrls_t *ctrl, int reqargc,
    */
 
   /* Check the 'down' ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "down")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "down")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -399,7 +416,7 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
   int res = 0;
 
   /* Check the kick ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "kick")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "kick")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -407,7 +424,8 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
   }
 
   /* Sanity check */
-  if (reqargc == 0 || reqargv == NULL) {
+  if (reqargc == 0 ||
+      reqargv == NULL) {
     pr_ctrls_add_response(ctrl, "missing required parameters");
     return -1;
   }
@@ -415,9 +433,30 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
   /* Handle 'kick user' requests. */
   if (strcmp(reqargv[0], "user") == 0) {
     register unsigned int i = 0;
+    int optc, kicked_count = 0, kicked_max = -1;
+    const char *reqopts = "n:";
     pr_scoreboard_entry_t *score = NULL;
 
-    if (reqargc == 1) {
+    pr_getopt_reset();
+
+    while ((optc = getopt(reqargc, reqargv, reqopts)) != -1) {
+      switch (optc) {
+        case 'n':
+          kicked_max = atoi(optarg);
+          if (kicked_max < 1) {
+            pr_ctrls_add_response(ctrl, "bad number: %s", optarg);
+            return -1;
+          }
+          break;
+
+        case '?':
+          pr_ctrls_add_response(ctrl, "unsupported option: '%c'",
+            (char) optopt);
+          return -1;
+      }
+    }
+
+    if (optind == reqargc) {
       pr_ctrls_add_response(ctrl, "kick user: missing required user name(s)");
       return -1;
     }
@@ -425,47 +464,99 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
     /* Iterate through the scoreboard, and send a SIGTERM to each
      * pid whose name matches the given user name(s).
      */
-    for (i = 1; i < reqargc; i++) {
+    for (i = optind; i < reqargc; i++) {
       unsigned char kicked_user = FALSE;
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+        pr_ctrls_add_response(ctrl, "error rewinding scoreboard: %s",
+          strerror(errno));
+        return -1;
+      }
 
       while ((score = pr_scoreboard_entry_read()) != NULL) {
+        pr_signals_handle();
+
+        if (kicked_max > 0 &&
+            kicked_count >= kicked_max) {
+          break;
+        }
+
         if (strcmp(reqargv[i], score->sce_user) == 0) {
+          int xerrno;
+
           res = 0;
 
           PRIVS_ROOT
           res = pr_scoreboard_entry_kill(score, SIGTERM);
+          xerrno = errno;
           PRIVS_RELINQUISH
 
-          if (res == 0)
+          if (res == 0) {
             kicked_user = TRUE;
+            kicked_count++;
 
-          else
-            ctrls_log("error kicking user '%s': %s", reqargv[i],
-              strerror(errno));
+          } else {
+            pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+              "error kicking user '%s': %s", reqargv[i], strerror(xerrno));
+          }
         }
       }
-      if (pr_restore_scoreboard() < 0)
-        ctrls_log("error restoring scoreboard: %s", strerror(errno));
+
+      if (pr_restore_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error restoring scoreboard: %s",
+          strerror(errno));
+      }
 
       if (kicked_user) {
-        pr_ctrls_add_response(ctrl, "kicked user '%s'", reqargv[i]);
-        ctrls_log("kicked user '%s'", reqargv[i]);
-        pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked user '%s'",
-          reqargv[i]);
+        if (kicked_max > 0) {
+          pr_ctrls_add_response(ctrl, "kicked user '%s' (%d clients)",
+            reqargv[i], kicked_max);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked user '%s' (%d clients)",
+            reqargv[i], kicked_max);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION
+            ": kicked user '%s' (%d clients)", reqargv[i], kicked_max);
 
-      } else
+        } else {
+          pr_ctrls_add_response(ctrl, "kicked user '%s'", reqargv[i]);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked user '%s'", reqargv[i]);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked user '%s'",
+            reqargv[i]);
+        }
+
+      } else {
         pr_ctrls_add_response(ctrl, "user '%s' not connected", reqargv[i]);
+      }
     }
 
   /* Handle 'kick host' requests. */
   } else if (strcmp(reqargv[0], "host") == 0) {
     register unsigned int i = 0;
+    int optc, kicked_count = 0, kicked_max = -1;
+    const char *reqopts = "n:";
     pr_scoreboard_entry_t *score = NULL;
 
-    if (reqargc == 1) {
+    pr_getopt_reset();
+
+    while ((optc = getopt(reqargc, reqargv, reqopts)) != -1) {
+      switch (optc) {
+        case 'n':
+          kicked_max = atoi(optarg);
+          if (kicked_max < 1) {
+            pr_ctrls_add_response(ctrl, "bad number: %s", optarg);
+            return -1;
+          }
+          break;
+
+        case '?':
+          pr_ctrls_add_response(ctrl, "unsupported option: '%c'",
+            (char) optopt);
+          return -1;
+      }
+    }
+
+    if (optind == reqargc) {
       pr_ctrls_add_response(ctrl, "kick host: missing required host(s)");
       return -1;
     }
@@ -475,49 +566,95 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
      * stringified IP address).
      */
 
-    for (i = 1; i < reqargc; i++) {
+    for (i = optind; i < reqargc; i++) {
       unsigned char kicked_host = FALSE;
       const char *addr;
       pr_netaddr_t *na;
 
-      na = pr_netaddr_get_addr(ctrl->ctrls_tmp_pool, reqargv[1], NULL);
-      if (!na) {
+      na = pr_netaddr_get_addr(ctrl->ctrls_tmp_pool, reqargv[i], NULL);
+      if (na == NULL) {
         pr_ctrls_add_response(ctrl, "kick host: error resolving '%s': %s",
-          reqargv[1], strerror(errno));
+          reqargv[i], strerror(errno));
         continue;
       }
 
       addr = pr_netaddr_get_ipstr(na);
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+        pr_ctrls_add_response(ctrl, "error rewinding scoreboard: %s",
+          strerror(errno));
+        return -1;
+      }
 
       while ((score = pr_scoreboard_entry_read()) != NULL) {
+        pr_signals_handle();
+
+        if (kicked_max > 0 &&
+            kicked_count >= kicked_max) {
+          break;
+        }
+
         if (strcmp(score->sce_client_addr, addr) == 0) {
           PRIVS_ROOT
-          if (pr_scoreboard_entry_kill(score, SIGTERM) == 0)
+          if (pr_scoreboard_entry_kill(score, SIGTERM) == 0) {
             kicked_host = TRUE;
+            kicked_count++;
+          }
           PRIVS_RELINQUISH
         }
       }
       pr_restore_scoreboard();
 
       if (kicked_host) {
-        pr_ctrls_add_response(ctrl, "kicked host '%s'", addr);
-        ctrls_log("kicked host '%s'", addr);
-        pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked host '%s'",
-          addr);
+        if (kicked_max > 0) {
+          pr_ctrls_add_response(ctrl, "kicked host '%s' (%d clients)", addr,
+            kicked_max);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked host '%s' (%d clients)",
+            addr, kicked_max);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION
+            ": kicked host '%s' (%d clients)", addr, kicked_max);
 
-      } else
+        } else {
+          pr_ctrls_add_response(ctrl, "kicked host '%s'", addr);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked host '%s'", addr);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked host '%s'",
+            addr);
+        }
+
+      } else {
         pr_ctrls_add_response(ctrl, "host '%s' not connected", addr);
+      }
     }
 
   /* Handle 'kick class' requests. */
   } else if (strcmp(reqargv[0], "class") == 0) {
     register unsigned int i = 0;
+    int optc, kicked_count = 0, kicked_max = -1;
+    const char *reqopts = "n:";
     pr_scoreboard_entry_t *score = NULL;
 
-    if (reqargc == 1) {
+    pr_getopt_reset();
+
+    while ((optc = getopt(reqargc, reqargv, reqopts)) != -1) {
+      switch (optc) {
+        case 'n':
+          kicked_max = atoi(optarg);
+          if (kicked_max < 1) {
+            pr_ctrls_add_response(ctrl, "bad client number: %s", optarg);
+            return -1;
+          }
+          break;
+
+        case '?':
+          pr_ctrls_add_response(ctrl, "unsupported option: '%c'",
+            (char) optopt);
+          return -1;
+      }
+    }
+
+    if (optind == reqargc) {
       pr_ctrls_add_response(ctrl, "kick class: missing required class name(s)");
       return -1;
     }
@@ -525,39 +662,71 @@ static int ctrls_handle_kick(pr_ctrls_t *ctrl, int reqargc,
     /* Iterate through the scoreboard, and send a SIGTERM to each
      * pid whose name matches the given class name(s).
      */
-    for (i = 1; i < reqargc; i++) {
+    for (i = optind; i < reqargc; i++) {
       unsigned char kicked_class = FALSE;
 
-      if (pr_rewind_scoreboard() < 0)
-        ctrls_log("error rewinding scoreboard: %s", strerror(errno));
+      if (pr_rewind_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error rewinding scoreboard: %s",
+          strerror(errno));
+        pr_ctrls_add_response(ctrl, "error rewinding scoreboard: %s",
+          strerror(errno));
+        return -1;
+      }
 
       while ((score = pr_scoreboard_entry_read()) != NULL) {
+        pr_signals_handle();
+
+        if (kicked_max > 0 &&
+            kicked_count >= kicked_max) {
+          break;
+        }
+
         if (strcmp(reqargv[i], score->sce_class) == 0) {
+          int xerrno;
+
           res = 0;
 
           PRIVS_ROOT
           res = pr_scoreboard_entry_kill(score, SIGTERM);
+          xerrno = errno;
           PRIVS_RELINQUISH
 
-          if (res == 0)
+          if (res == 0) {
             kicked_class = TRUE;
+            kicked_count++;
 
-          else
-            ctrls_log("error kicking class '%s': %s", reqargv[i],
-              strerror(errno));
+          } else {
+            pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+              "error kicking class '%s': %s", reqargv[i], strerror(xerrno));
+          }
         }
       }
-      if (pr_restore_scoreboard() < 0)
-        ctrls_log("error restoring scoreboard: %s", strerror(errno));
+
+      if (pr_restore_scoreboard() < 0) {
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "error restoring scoreboard: %s",
+          strerror(errno));
+      }
 
       if (kicked_class) {
-        pr_ctrls_add_response(ctrl, "kicked class '%s'", reqargv[i]);
-        ctrls_log("kicked class '%s'", reqargv[i]);
-        pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked class '%s'",
-          reqargv[i]);
+        if (kicked_max > 0) {
+          pr_ctrls_add_response(ctrl, "kicked class '%s' (%d clients)",
+            reqargv[i], kicked_max);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+            "kicked class '%s' (%d clients)", reqargv[i], kicked_max);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION
+            ": kicked class '%s' (%d clients)", reqargv[i], kicked_max);
 
-      } else
+        } else {
+          pr_ctrls_add_response(ctrl, "kicked class '%s'", reqargv[i]);
+          pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "kicked class '%s'",
+            reqargv[i]);
+          pr_log_debug(DEBUG4, MOD_CTRLS_ADMIN_VERSION ": kicked class '%s'",
+            reqargv[i]);
+        }
+
+      } else {
         pr_ctrls_add_response(ctrl, "class '%s' not connected", reqargv[i]);
+      }
     }
 
   } else {
@@ -573,7 +742,7 @@ static int ctrls_handle_restart(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
 
   /* Check the restart ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "restart")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "restart")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -581,16 +750,35 @@ static int ctrls_handle_restart(pr_ctrls_t *ctrl, int reqargc,
   }
 
   /* Be pedantic */
-  if (reqargc != 0) {
+  if (reqargc > 1) {
     pr_ctrls_add_response(ctrl, "bad number of arguments");
     return -1;
   }
 
-  PRIVS_ROOT
-  raise(SIGHUP);
-  PRIVS_RELINQUISH
+  if (reqargc == 0) {
+    PRIVS_ROOT
+    raise(SIGHUP);
+    PRIVS_RELINQUISH
+
+    pr_ctrls_add_response(ctrl, "restarted server");
+
+  } else if (reqargc == 1) {
+    if (strcmp(reqargv[0], "count") == 0) {
+      struct tm *tm;
+
+      tm = pr_gmtime(ctrl->ctrls_tmp_pool, &ctrls_admin_start);
+      pr_ctrls_add_response(ctrl,
+        "server restarted %u %s since %04d-%02d-%02d %02d:%02d:%02d GMT",
+        ctrls_admin_nrestarts, ctrls_admin_nrestarts != 1 ? "times" : "time",
+        tm->tm_year+1900, tm->tm_mon+1, tm->tm_mday, tm->tm_hour, tm->tm_min,
+        tm->tm_sec);
+
+    } else {
+      pr_ctrls_add_response(ctrl, "unsupported parameter '%s'", reqargv[0]);
+      return -1;
+    }
+  }
 
-  pr_ctrls_add_response(ctrl, "restarted server");
   return 0;
 }
 
@@ -598,7 +786,7 @@ static int ctrls_handle_scoreboard(pr_ctrls_t *ctrl, int reqargc,
     char **reqargv) {
 
   /* Check the scoreboard ACL. */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "scoreboard")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "scoreboard")) {
 
     /* Access denied. */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -629,7 +817,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
   char **respargv = NULL;
 
   /* Check the shutdown ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "shutdown")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "shutdown")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -650,7 +838,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
       timeout = atoi(reqargv[1]);
       time(&now);
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
         "shutdown: waiting %u seconds before shutting down", timeout);
 
       /* If the timeout is less than the waiting period, reduce the
@@ -673,7 +861,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
       if (timeout &&
           time(NULL) - now > timeout) {
 
-        ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+        pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
           "shutdown: %u seconds elapsed, ending %lu remaining sessions",
           timeout, nkids);
 
@@ -685,7 +873,7 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
         break;
       }
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
         "shutdown: waiting for %lu sessions to end", nkids);
       sleep(waiting);
 
@@ -709,17 +897,18 @@ static int ctrls_handle_shutdown(pr_ctrls_t *ctrl, int reqargc,
   /* Manually tweak the return value, for the benefit of the client */
   ctrl->ctrls_cb_retval = 0;
 
-  if (pr_ctrls_flush_response(ctrl) < 0)
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+  if (pr_ctrls_flush_response(ctrl) < 0) {
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
       "shutdown: error flushing response: %s", strerror(errno));
+  }
 
   /* For logging/accounting purposes */
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
     "shutdown: flushed to %s/%s client: return value: 0",
     ctrl->ctrls_cl->cl_user, ctrl->ctrls_cl->cl_group);
 
   for (i = 0; i < respargc; i++)
-    ctrls_log(MOD_CTRLS_ADMIN_VERSION,
+    pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION,
       "shutdown: flushed to %s/%s client: '%s'",
       ctrl->ctrls_cl->cl_user, ctrl->ctrls_cl->cl_group, respargv[i]);
 
@@ -733,7 +922,7 @@ static int admin_addr_status(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
     unsigned int port) {
   pr_ipbind_t *ipbind = NULL;
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   /* Fetch the ipbind associated with this address/port. */
@@ -756,7 +945,7 @@ static int ctrls_handle_status(pr_ctrls_t *ctrl, int reqargc,
   register unsigned int i = 0;
 
   /* Check the status ACL. */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "status")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "status")) {
 
     /* Access denied. */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -779,7 +968,7 @@ static int ctrls_handle_status(pr_ctrls_t *ctrl, int reqargc,
     if (strcasecmp(server_str, "all") == 0) {
       pr_ipbind_t *ipbind = NULL;
 
-      ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking all servers");
+      pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "status: checking all servers");
 
       while ((ipbind = pr_ipbind_get(ipbind)) != NULL) {
         const char *ipbind_str = pr_netaddr_get_ipstr(ipbind->ib_addr); 
@@ -825,7 +1014,7 @@ static int ctrls_handle_trace(pr_ctrls_t *ctrl, int reqargc,
 #ifdef PR_USE_TRACE
 
   /* Check the trace ACL. */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "trace")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "trace")) {
 
     /* Access denied. */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -843,10 +1032,10 @@ static int ctrls_handle_trace(pr_ctrls_t *ctrl, int reqargc,
 
     for (i = 0; i < reqargc; i++) {
       char *channel, *tmp;
-      int level;
+      int min_level, max_level, res;
 
       tmp = strchr(reqargv[i], ':');
-      if (!tmp) {
+      if (tmp == NULL) {
         pr_ctrls_add_response(ctrl, "trace: badly formatted parameter: '%s'",
           reqargv[i]);
         return -1;
@@ -854,17 +1043,25 @@ static int ctrls_handle_trace(pr_ctrls_t *ctrl, int reqargc,
 
       channel = reqargv[i];
       *tmp = '\0';
-      level = atoi(++tmp);
 
-      if (pr_trace_set_level(channel, level) < 0) {
-        pr_ctrls_add_response(ctrl,
-          "trace: error setting channel '%s' to level %d: %s", channel, level,
-          strerror(errno));
-        return -1;
+      res = pr_trace_parse_levels(tmp + 1, &min_level, &max_level);
+      if (res == 0) {
+        if (pr_trace_set_levels(channel, min_level, max_level) < 0) {
+          pr_ctrls_add_response(ctrl,
+            "trace: error setting channel '%s' to levels %d-%d: %s", channel,
+            min_level, max_level, strerror(errno));
+          return -1;
+
+        } else {
+          pr_ctrls_add_response(ctrl, "trace: set channel '%s' to levels %d-%d",
+            channel, min_level, max_level);
+        }
 
       } else {
-        pr_ctrls_add_response(ctrl, "trace: set channel '%s' to level %d",
-          channel, level);
+        pr_ctrls_add_response(ctrl,
+          "trace: error parsing level '%s' for channel '%s': %s", tmp + 1,
+          channel, strerror(errno));
+        return -1;
       }
     }
  
@@ -931,27 +1128,27 @@ static int admin_addr_up(pr_ctrls_t *ctrl, pr_netaddr_t *addr,
    * share the same listen connection.
    */
   if (ipbind->ib_server->ServerPort && !ipbind->ib_server->listen) {
-    ipbind->ib_server->listen =
-      pr_inet_create_connection(ipbind->ib_server->pool, server_list, -1,
+    ipbind->ib_server->listen = pr_ipbind_get_listening_conn(ipbind->ib_server,
       (SocketBindTight ? ipbind->ib_server->addr : NULL),
-      ipbind->ib_server->ServerPort, FALSE);
+      ipbind->ib_server->ServerPort);
   }
 
-  ctrls_log(MOD_CTRLS_ADMIN_VERSION, "up: attempting to enable %s#%u",
+  pr_ctrls_log(MOD_CTRLS_ADMIN_VERSION, "up: attempting to enable %s#%u",
     pr_netaddr_get_ipstr(addr), port);
 
   PR_OPEN_IPBIND(ipbind->ib_server->addr, ipbind->ib_server->ServerPort,
     ipbind->ib_server->listen, FALSE, FALSE, TRUE);
 
-  if (res < 0)
+  if (res < 0) {
     pr_ctrls_add_response(ctrl, "up: no server listening on %s#%u",
       pr_netaddr_get_ipstr(addr), port);
-  else
+
+  } else {
     pr_ctrls_add_response(ctrl, "up: %s#%u enabled",
       pr_netaddr_get_ipstr(addr), port);
+  }
 
   PR_ADD_IPBINDS(ipbind->ib_server);
-
   return 0;
 }
 
@@ -964,7 +1161,7 @@ static int ctrls_handle_up(pr_ctrls_t *ctrl, int reqargc,
    */
 
   /* Check the 'up' ACL */
-  if (!ctrls_check_acl(ctrl, ctrls_admin_acttab, "up")) {
+  if (!pr_ctrls_check_acl(ctrl, ctrls_admin_acttab, "up")) {
 
     /* Access denied */
     pr_ctrls_add_response(ctrl, "access denied");
@@ -1037,7 +1234,7 @@ MODRET set_adminctrlsacls(cmd_rec *cmd) {
       strcmp(cmd->argv[3], "group") != 0)
     CONF_ERROR(cmd, "third parameter must be 'user' or 'group'");
 
-  bad_action = ctrls_set_module_acls(ctrls_admin_acttab, ctrls_admin_pool,
+  bad_action = pr_ctrls_set_module_acls(ctrls_admin_acttab, ctrls_admin_pool,
     actions, cmd->argv[2], cmd->argv[3], cmd->argv[4]);
   if (bad_action != NULL)
     CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown action: '",
@@ -1080,7 +1277,7 @@ MODRET set_adminctrlsengine(cmd_rec *cmd) {
      */
     char **actions = ctrls_parse_acl(cmd->tmp_pool, cmd->argv[1]);
 
-    bad_action = ctrls_unregister_module_actions(ctrls_admin_acttab, actions,
+    bad_action = pr_ctrls_unregister_module_actions(ctrls_admin_acttab, actions,
       &ctrls_admin_module);
     if (bad_action != NULL)
       CONF_ERROR(cmd, pstrcat(cmd->tmp_pool, ": unknown action: '",
@@ -1129,9 +1326,10 @@ static void ctrls_admin_restart_ev(const void *event_data, void *user_data) {
     /* Allocate and initialize the ACL for this control. */
     ctrls_admin_acttab[i].act_acl = pcalloc(ctrls_admin_pool,
       sizeof(ctrls_acl_t));
-    ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
+    pr_ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
   }
 
+  ctrls_admin_nrestarts++;
   return;
 }
 
@@ -1182,12 +1380,12 @@ static int ctrls_admin_init(void) {
     /* Allocate and initialize the ACL for this control. */
     ctrls_admin_acttab[i].act_acl = pcalloc(ctrls_admin_pool,
       sizeof(ctrls_acl_t));
-    ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
+    pr_ctrls_init_acl(ctrls_admin_acttab[i].act_acl);
 
     if (pr_ctrls_register(&ctrls_admin_module,
         ctrls_admin_acttab[i].act_action, ctrls_admin_acttab[i].act_desc,
         ctrls_admin_acttab[i].act_cb) < 0)
-     pr_log_pri(PR_LOG_INFO, MOD_CTRLS_ADMIN_VERSION
+     pr_log_pri(PR_LOG_NOTICE, MOD_CTRLS_ADMIN_VERSION
         ": error registering '%s' control: %s",
         ctrls_admin_acttab[i].act_action, strerror(errno));
   }
@@ -1201,6 +1399,7 @@ static int ctrls_admin_init(void) {
   pr_event_register(&ctrls_admin_module, "core.startup",
     ctrls_admin_startup_ev, NULL);
 
+  time(&ctrls_admin_start);
   return 0;
 }
 
