@@ -10,9 +10,9 @@ my $conn_ex;
 
 sub new {
   my $class = shift;
-  my ($addr, $port, $use_pasv, $timeout) = @_;
-  $use_pasv = 0 unless defined($use_pasv);
-  $timeout = 10 unless defined($timeout);
+  my ($addr, $port, $use_port, $conn_timeout, $cmd_timeout) = @_;
+  $use_port = 0 unless defined($use_port);
+  $conn_timeout = 10 unless defined($conn_timeout);
  
   my $ftp;
 
@@ -24,23 +24,27 @@ sub new {
   # given timeout.
 
   my %opts = (
-      Port => $port,
+    Port => $port,
   );
 
-  if ($use_pasv) {
-    $opts{Passive} = 1;
+  if ($use_port) {
+    $opts{Passive} = 0;
 
   } else {
-    $opts{Passive} = 0;
+    $opts{Passive} = 1;
   }
 
   if ($ENV{TEST_VERBOSE}) {
     $opts{Debug} = 10;
   }
 
+  if (defined($cmd_timeout)) {
+    $opts{Timeout} = $cmd_timeout;
+  }
+
   while (1) {
-    if (time() - $now > $timeout) {
-      croak("Unable to connect to $addr:$port: Timed out after $timeout secs");
+    if (time() - $now > $conn_timeout) {
+      croak("Unable to connect to $addr:$port: Timed out after $conn_timeout secs");
     }
 
     $ftp = Net::FTP->new($addr, %opts);
@@ -53,6 +57,16 @@ sub new {
     sleep(1);
   }
 
+  # Naughtily invade the Net::FTP internals; it makes for less confusion
+  # when writing the unit tests.
+  if (exists($ENV{FTP_FIREWALL})) {
+    ${*$ftp}{net_ftp_firewall} = $ENV{FTP_FIREWALL};
+  }
+
+  if (exists($ENV{FTP_FIREWALL_TYPE})) {
+    ${*$ftp}{net_ftp_firewall_type} = $ENV{FTP_FIREWALL_TYPE};
+  }
+
   my $self = {
     addr => $addr,
     ftp => $ftp,
@@ -72,8 +86,12 @@ sub response_code {
 
 sub response_msg {
   my $self = shift;
-  my $index = shift;
-  $index = 1 unless defined($index);
+  my $req_index = shift;
+
+  my $index = 1;
+  if (defined($req_index)) {
+    $index = $req_index;
+  }
 
   if (defined($self->{mesg})) {
     my $msg = $self->{mesg};
@@ -83,12 +101,27 @@ sub response_msg {
   }
 
   my @msgs = $self->{ftp}->message;
-  if (scalar(@msgs) > 1) {
+  my $nmsgs = scalar(@msgs);
+  if ($nmsgs > 1) {
+    if ($index > ($nmsgs - 1)) {
+      return undef;
+    }
+
     chomp($msgs[$index]);
     return $msgs[$index];
+
+  } else {
+    if (defined($req_index)) {
+      if ($index > 0) {
+        return undef;
+      }
+    }
+  }
+
+  if (defined($msgs[0])) {
+    chomp($msgs[0]);
   }
 
-  chomp($msgs[0]);
   return $msgs[0];
 }
 
@@ -151,7 +184,22 @@ sub login {
   $SIG{ALRM} = \&login_alarm;
   alarm($login_timeout);
 
+  # Work around some (strange? broken?) ness in Net::FTP's handling of
+  # the destination server in the login() method for "firewalls"
+  # (i.e. proxying).
+
+  my $ftp = $self->{ftp};
+  my $net_ftp_host = ${*$ftp}{net_ftp_host};
+
+  if (exists($ENV{FTP_FIREWALL})) {
+    ${*$ftp}{net_ftp_host} = $ENV{FTP_FIREWALL};
+  }
+
   unless ($self->{ftp}->login($user, $pass)) {
+    if (exists($ENV{FTP_FIREWALL})) {
+      ${*$ftp}{net_ftp_host} = $net_ftp_host;
+    }
+
     alarm(0);
     $SIG{ALRM} = 'DEFAULT';
 
@@ -618,6 +666,29 @@ sub port {
   # Naughtily invade the Net::FTP internals; it makes for less confusion
   # when writing the unit tests.
   my $ftp = $self->{ftp};
+
+  if ($port) {
+    # Determine the local port from the given argument.
+
+    my $numbers = [split(',', $port)];
+    my $local_port = ($numbers->[4] * 256) + $numbers->[5];
+
+    # If the caller provided an explicit PORT argument, then we need to
+    # open the listening socket ourselves.  Net::FTP is braindead that way.
+    #
+    # The code below is copied from Net::FTP::port().
+
+    ${*$ftp}{net_ftp_listen} ||= IO::Socket::INET->new(
+      Listen => 5,
+      Proto => 'tcp',
+      Timeout => $ftp->timeout,
+      LocalAddr => $ftp->sockhost,
+      LocalPort => $local_port,
+    );
+
+    ${*$ftp}{net_ftp_intern_port} = 1;
+  }
+
   delete(${*$ftp}{net_ftp_passive});
 
   my $msg = $self->response_msg();
@@ -1313,6 +1384,32 @@ sub site {
   }
 }
 
+sub quote {
+  my $self = shift;
+  my $cmd = shift;
+  $cmd = '' unless defined($cmd);
+  my $code;
+
+  $code = $self->{ftp}->quot($cmd, @_);
+  unless ($code) {
+    croak("Raw command '$cmd' failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  if ($code == 4 || $code == 5) {
+    croak("Raw command '$cmd' failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
 sub mlsd {
   my $self = shift;
   my $path = shift;
@@ -1507,4 +1604,74 @@ sub get_connect_exception {
   return $conn_ex;
 }
 
+sub stat {
+  my $self = shift;
+  my $path = shift;
+  $path = '' unless defined($path);
+  my $code;
+
+  $code = $self->{ftp}->quot('STAT', $path);
+  unless ($code) {
+    croak("STAT command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  if ($code == 4 || $code == 5) {
+    croak("STAT command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
+# From the FTP HOST command RFCXXXX (currently in Draft form)
+sub host {
+  my $self = shift;
+  my $host = shift;
+  $host = '' unless defined($host);
+  my $code;
+
+  $code = $self->{ftp}->quot('HOST', $host);
+  unless ($code) {
+    croak("HOST command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  if ($code == 4 || $code == 5) {
+    croak("HOST command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
+sub abort {
+  my $self = shift;
+
+  unless ($self->{ftp}->abort()) {
+    croak("ABOR command failed: " .  $self->{ftp}->code . ' ' .
+      $self->response_msg());
+  }
+
+  my $msg = $self->response_msg();
+  if (wantarray()) {
+    return ($self->{ftp}->code, $msg);
+
+  } else {
+    return $msg;
+  }
+}
+
 1;
